import pandas as pd

from datetime import datetime
from pathlib import Path

from nsys_recipe.lib import helpers
from nsys_recipe.lib import loader
from nsys_recipe.lib import nvtx
from nsys_recipe.lib import pace
from nsys_recipe.lib import recipe
from nsys_recipe.lib import summary

from nsys_recipe.lib.args import Option
from nsys_recipe.lib.loader import TableConfig

class NvtxPace(recipe.Recipe):
    @staticmethod
    def mapper_func(nsysrep, parsed_args):
        sqlite_file = helpers.nsysrep_to_sqlite(nsysrep)
        if sqlite_file is None:
            return None

        name_column = 'text'
        table_config_dict = {
            'TARGET_INFO_SESSION_START_TIME': TableConfig(),
            'StringIds': TableConfig(),
            'NVTX_EVENTS': TableConfig(
                [name_column, 'start', 'end', 'textId', 'globalTid', 'endGlobalTid'],
                time_filter=True
            )
        }

        df_dict = loader.read_sqlite_tables(
            sqlite_file,
            table_config_dict,
            parsed_args.start,
            parsed_args.end
        )
        if df_dict is None:
            return None

        range_df = nvtx.combine_text_fields(df_dict['NVTX_EVENTS'], df_dict['StringIds'])
        range_df = pace.filter_by_pace_name(range_df, name_column, parsed_args.name)
        if range_df.empty:
            print(f"{sqlite_file} does not contain '{parsed_args.name}'.")
            return None

        filename = Path(sqlite_file).stem
        session_start = pace.get_session_start_time(df_dict['TARGET_INFO_SESSION_START_TIME'])
        pace_df, stats_df = pace.compute_pace_stats_dfs(range_df, name_column)
        return pace.PaceInfo(filename, pace_df, stats_df, session_start)

    def reducer_func(self, mapper_res):
        pace_infos = helpers.filter_none(mapper_res)
        pace_infos = sorted(pace_infos, key=lambda x: x.filename)

        filenames, pace_dfs, stats_dfs, session_starts = zip(*pace_infos)
        pace.apply_time_offset(session_starts, pace_dfs)

        rank_file_df = pd.DataFrame(filenames, columns=['File'])
        rank_file_df.to_parquet(self.add_output_file('files.parquet'), index=False)

        stats_df = summary.aggregate_stats_dfs(stats_dfs)
        stats_df.to_parquet(self.add_output_file('stats.parquet'))

        for name, df in pace.split_columns_as_dataframes(pace_dfs).items():
            df.to_parquet(self.add_output_file(f'pace_{name}.parquet'), index=False)

    def save_notebook(self):
        self.create_notebook('pace.ipynb')
        self.add_notebook_helper_file('nsys_pres.py')

    def save_analysis_file(self):
        self._analysis_dict.update({
            'EndTime': str(datetime.now()),
            'InputReports': self._parsed_args.input,
            'Outputs': self._output_files
        })
        self.create_analysis_file()

    def run(self, context):
        super().run(context)

        mapper_res = context.wait(context.map(
            self.mapper_func,
            self._parsed_args.input,
            parsed_args=self._parsed_args
        ))
        self.reducer_func(mapper_res)

        self.save_notebook()
        self.save_analysis_file()

    @classmethod
    def get_argument_parser(cls):
        parser = super().get_argument_parser()

        mutually_exclusive_group = parser.recipe_group.add_mutually_exclusive_group(required=True)
        parser.add_argument_to_group(mutually_exclusive_group, Option.REPORT_DIR)
        parser.add_argument_to_group(mutually_exclusive_group, Option.INPUT)

        parser.add_recipe_argument(Option.OUTPUT)
        parser.add_recipe_argument(Option.FORCE_OVERWRITE)
        parser.add_recipe_argument(Option.AUTO_INCREMENT)
        parser.add_recipe_argument(
            '--name',
            type=str,
            help="Name of the NVTX range used as delineator between iterations",
            required=True)
        parser.add_recipe_argument(Option.START)
        parser.add_recipe_argument(Option.END)

        return parser
